from enum import Enum
from typing import Optional, Union, Dict
import re
from functools import total_ordering


class Currency(Enum):
    USD = "$"
    GBP = "£"
    EUR = "€"
    TWD = "NT$"


@total_ordering
class Price:
    """
    A monetary price, including a currency and a positive amount.
    Providing a negative amount will result in an error (note that we only allow
    positive amounts for safety reasons, if you are really in need of representing
    negative amounts depending on your use case you should associate some meta info to
    the price, for example for cash transactions this is done with CASH_OUT vs CASH_IN)
    """
    def __init__(self, currency: Currency, amount: float):
        self.currency = currency
        self.amount = amount

    @property
    def amount(self) -> float:
        return self._amount

    @amount.setter
    def amount(self, amount: float):
        if amount < 0:
            raise ValueError(("Price can only be positive"))
        self._amount = abs(amount)
    
    @staticmethod
    def from_str(price: str, currency: Optional[Currency] = None, expect_negative=False) -> "Price":
        match = re.search(r"[\d\.,]+", price)
        if match is None:
            raise ValueError(f'Cannot find valid price in string "{price}"')
        amount = float(match.group().replace(",", ""))
        if price.strip().startswith("-"):
            amount = -amount

        if currency is None:
            for known_currency in Currency:
                for symbol in [known_currency.name, known_currency.value]:
                    if symbol in price:
                        currency = known_currency
                        break

                if currency is not None:
                    break

            if currency is None:
                raise ValueError(f"Couldn't get currency for price {str}")

        if expect_negative:
            if amount > 0:
                raise ValueError(f"Expected a negative value")
            amount = - amount
        return Price(currency=currency, amount=amount)

    def normalize_currency(self, conversion_rates: Dict[str,float]) -> float:
        """ Convert price to dest currency """
        rate = conversion_rates.get(self.currency.name)
        if rate is None:
            raise ValueError(f"Unsupported currency {self.currency.name}")
        return self.amount / rate

    def to_string(self):
        return f"{self.currency.value}{self.amount :,.2f}"

    def __repr__(self):
        return self.to_string()

    def __add__(self, other: Union["Price", float]):
        if isinstance(other, float):
            return Price(currency=self.currency, amount=self.amount + other)
        elif isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError("Prices have different currencies")
            return Price(currency=self.currency, amount=self.amount + other.amount)
        raise ValueError(f"Incompatible type {type(other)}")

    def __sub__(self, other: Union["Price", float]):
        if isinstance(other, float):
            return Price(currency=self.currency, amount=self.amount - other)
        elif isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError("Prices have different currencies")
            return Price(currency=self.currency, amount=self.amount - other.amount)
        raise ValueError(f"Incompatible type {type(other)}")

    def __mul__(self, other: Union["Price", float]):
        if isinstance(other, float):
            return Price(currency=self.currency, amount=self.amount * other)
        elif isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError("Prices have different currencies")
            return Price(currency=self.currency, amount=self.amount * other.amount)
        raise ValueError(f"Incompatible type {type(other)}")

    def __eq__(self, other: Union["Price", float]):
        if isinstance(other, float):
            return self.amount == other
        elif isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError("Prices have different currencies")
            return self.amount == other.amount
        raise ValueError(f"Incompatible type {type(other)}")

    def __lt__(self, other: Union["Price", float]):
        if isinstance(other, float):
            return self.amount < other
        elif isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError("Prices have different currencies")
            return self.amount < other.amount
        raise ValueError(f"Incompatible type {type(other)}")



from enum import Enum
from typing import Optional
from datetime import datetime
from .price import Price


class TransactionAction(Enum):
    BUY = "buy"
    SELL = "sell"
    VEST = "vest"
    TAX = "tax"
    DIV = "dividends"
    CASH_IN = "cash_in"
    CASH_OUT = "cash_out"
    SPLIT = "split"

    @staticmethod
    def from_schwab_action(action: str) -> "TransactionAction":
        action = action.lower()
        if action == "buy" or action == "reinvest shares":
            return TransactionAction.BUY
        elif action == "sell":
            return TransactionAction.SELL
        elif action == "stock plan activity":
            return TransactionAction.VEST
        elif action == "nra tax adj":
            return TransactionAction.TAX
        elif action == "foreign tax paid":
            return TransactionAction.TAX
        elif action == "qual div reinvest":
            return TransactionAction.DIV
        elif action == "qualified dividend":
            return TransactionAction.DIV
        raise ValueError(f'Cannot translate action "{action}"')


class Transaction:
    def __init__(
        self,
        action: TransactionAction,
        date: datetime,
        quantity: Optional[float] = None,
        price: Optional[Price] = None,
        ticker: Optional[str] = None,
        isin: Optional[str] = None,
        name: Optional[str] = None,
        fees: Optional[Price] = None,
        amount: Optional[Price] = None,
    ):
        self.action = action
        self.date = date
        self.quantity = None if quantity is None else abs(quantity)
        self.price = price
        self.fees = fees
        self.name = name
        self.ticker = ticker
        self.isin = isin
        self.amount = amount

        self._check_constraints()
        self._infer_missing_fields()

    def __str__(self):
        return f"Transaction: {self.date} {self.action.value} {self.name} {self.amount}"

    def get_holding_key(self):
        if (
            self.action != TransactionAction.BUY
            and self.action != TransactionAction.SELL
            and self.action != TransactionAction.VEST
            and self.action != TransactionAction.SPLIT
        ):
            raise ValueError(
                f"Cannot get holding key from transaction with action {self.action}"
            )

        # prioritize stronger identifiers
        if self.isin is not None:
            return self.isin
        elif self.ticker is not None:
            return self.ticker
        elif self.name is not None:
            return self.name
        raise ValueError("Cannot construct holding name")

    def _infer_missing_fields(self):
        if self.amount is None and self.quantity is not None and self.price is not None:
            fees = 0. if self.fees is None else self.fees
            self.amount = Price(
                currency=self.price.currency, amount=(self.quantity * self.price.amount) - fees
            )

    def _check_constraints(self):
        if self.quantity is not None and self.quantity <= 0:
            raise ValueError("Quantity must be > 0")

        if self.price is not None and self.price.amount <= 0:
            raise ValueError("Price must be > 0")

        if self.fees is not None and self.fees.amount <= 0:
            raise ValueError("Fees must be > 0")

        currency = None
        for price_arg in [self.price, self.fees, self.amount]:
            if price_arg is not None:
                if currency is None:
                    currency = price_arg.currency
                elif currency != price_arg.currency:
                    raise ValueError(
                        "Currency of price, fees and amount must be the same"
                    )

        if self.action == TransactionAction.BUY:
            self._check_constraints_buy()
        elif self.action == TransactionAction.SELL:
            self._check_constraints_sell()
        elif self.action == TransactionAction.VEST:
            self._check_constraints_vest()
        elif self.action == TransactionAction.TAX:
            self._check_constraints_tax()
        elif self.action == TransactionAction.CASH_IN:
            self._check_constraints_cash()
        elif self.action == TransactionAction.CASH_OUT:
            self._check_constraints_cash()
        elif self.action == TransactionAction.DIV:
            self._check_constraints_div()

    def _check_has_identifier(self):
        if not any([self.name, self.ticker, self.isin]):
            raise ValueError("Didn't find any valid identifier for transaction")

    def _check_constraints_buy(self):
        # Quantity must be set, default to one
        if self.quantity is None:
            self.quantity = 1

        # Buy transactions must have an identifier
        self._check_has_identifier()

        # Price must be set, or we should be able to fetch it
        if self.price is None and self.amount is None:
            raise ValueError("Price or amount must be set for buy/sell transactions")

        # Infer price from amount
        if self.price is None:
            self.price = Price(
                currency=self.amount.currency,
                amount=abs(self.amount.amount / self.quantity),
            )

        # Currency must be consistent
        if self.amount is not None and self.price is not None:
            if self.amount.currency != self.price.currency:
                raise ValueError("Price and amount must use the same currency")


    def _check_constraints_sell(self):
        # Same constrains as buy transactions
        self._check_constraints_buy()

    def _check_constraints_vest(self):
        if self.quantity is None:
            raise ValueError("Quantity must be set for vest transactions")

    def _check_constraints_tax(self):
        if self.amount is None:
            raise ValueError("Amount must be set for tax transactions")

    def _check_constraints_cash(self):
        if self.amount is None:
            raise ValueError("Amount must be set for cash transactions")

    def _check_constraints_div(self):
        if self.amount is None:
            raise ValueError("Amount must be set for dividends transactions")

    def __eq__(self, o):
        return (
            self.action == o.action and
            self.date == o.date and
            self.ticker == o.ticker and
            self.isin == o.isin and
            self.name == o.name and
            self.fees == o.fees and
            self.price == o.price and
            self.quantity == o.quantity and
            self.amount == o.amount
        )



from typing import Dict, List, Iterable, Optional
from collections import defaultdict
import pandas as pd
import numpy as np
from .transaction import TransactionAction, Transaction
from .balance import Balance
from .price import Price, Currency


class Analysis:
    def __init__(
        self,
        prices: pd.DataFrame,
        conv_rates: Dict[str, float],
        holdings_currencies: Dict[str, Currency],
    ):
        self.prices = prices
        self.conv_rates = conv_rates
        self.holdings_currencies = holdings_currencies
        self.holdings_keys = list(self.prices.columns)

    def get_allocations(
        self, balances: Iterable[Balance], ndays: Optional[int] = None
    ) -> pd.DataFrame:
        """
        Given a list of known balances, creates a pandas dataframe of allocations
        where each column represent an holding and each row represent holdings'
        values (i.e.  as cash) for a day.
        An extra cash column is added to represent the amount of cash held.

        For example, considering a single day X, two holdings which corresponding
        value in cash in 100$ each and 50$ cash deposited in the account, then the
        resulting allocations dataframe will look like this:

                    | HOLDING_1 | HOLDING_2 | cash
                ----+-----------+-----------+--------
                X   |   100     |   100     | 50

        If in day X+1 we have bought 50$ worth of HOLDING_1 and HOLDING_2's price
        has increased +5%, then the allocation table will look like this:

                    | HOLDING_1 | HOLDING_2 | cash
                ----+-----------+-----------+--------
                X   |   100     |   100     | 50
                X+1 |   150     |   105     |  0
        """

        # New dataframe with the same layout of prices
        allocations = pd.DataFrame().reindex_like(self.prices)
        cash = pd.DataFrame(columns=["cash"], index=allocations.index, dtype=np.float64)

        for balance in balances:
            b_date = pd.Timestamp(balance.date)

            # Skip balances out of bounds
            if b_date > allocations.index.max() or b_date < allocations.index.min():
                continue

            # Populate holdings columns
            for holding in balance.holdings.values():
                allocations.loc[
                    pd.Timestamp(balance.date), holding.get_key()
                ] = holding.quantity

            # Populate cash dataframe
            total_cash = 0.0
            for curr, amount in balance.cash.items():
                cash_to_add = Price(
                    currency=curr, amount=abs(amount)
                ).normalize_currency(self.conv_rates)
                # Prices are always positive, so must compensate for that
                total_cash += cash_to_add if amount > 0 else -cash_to_add
            cash.loc[pd.Timestamp(balance.date), "cash"] = total_cash

        # Fill empty slot using previous known values
        allocations = allocations.interpolate(method="pad", axis=0)
        cash = cash.interpolate(method="pad", axis=0)

        # Apply prices
        allocations *= self.prices

        # For each column, apply conversion rate
        for key in self.prices.columns:
            currency = self.holdings_currencies.get(key)
            if currency is None:
                raise ValueError(f"Couldn't determine currency for holding {key}")

            rate = self.conv_rates.get(currency.name)
            if rate is None:
                raise ValueError(f"Unsupported currency {currency.name}")

            allocations[key] /= rate

        # Return holdings allocations including cash
        res = pd.concat([allocations, cash], axis=1).fillna(0)

        # Take only the last n days
        if ndays is not None:
            # pylint: disable=invalid-unary-operand-type
            res = res.iloc[-ndays:]

        return res

    def get_earnings(
        self,
        allocations: pd.DataFrame,
        transactions: List[Transaction],
        ndays: Optional[int] = None,
    ) -> pd.Series:
        """
        Computes net earnings by discounting incoming cash flows that are not
        directly coming from an investment (e.g.  cash transfer, vesting).
        """
        earnings = allocations.sum(axis=1)

        for trs in transactions:
            # Discount cash put into the account
            if (
                trs.action == TransactionAction.CASH_IN
                or trs.action == TransactionAction.CASH_OUT
            ):
                rows = earnings.loc[trs.date :]
                delta = trs.amount.normalize_currency(self.conv_rates)
                earnings.loc[trs.date :] = rows.add(
                    -delta if trs.action == TransactionAction.CASH_IN else delta
                )

            # Discount vested stock (as it is not earning from investiment)
            elif trs.action == TransactionAction.VEST:
                rows = earnings.loc[trs.date :]
                price = self.prices.loc[trs.date :][trs.get_holding_key()].iloc[0]
                holding_cur = self.holdings_currencies[trs.get_holding_key()]
                delta = Price(
                    currency=holding_cur, amount=trs.quantity * price
                ).normalize_currency(self.conv_rates)
                earnings.loc[trs.date :] = rows.add(-delta)

        # Take only the last n days
        if ndays is not None:
            # pylint: disable=invalid-unary-operand-type
            earnings = earnings.iloc[-ndays:]

        # Subtract initial value so that we always start from zero
        earnings = earnings.add(-earnings.iloc[0])

        return earnings

    def get_ror(
        self,
        allocations: pd.DataFrame,
        earnings: pd.Series,
    ) -> float:
        """
        Calculates time weighted rate of return (TWROR)
        https://www.ironsidegroup.com/2013/01/01/calculating-the-time-weighted-rate-of-return-in-a-cognos-report/
        """
        deltas = earnings.shift(-1) - earnings
        balances = allocations.sum(axis=1)
        balances = (balances + deltas) / balances
        ret = balances.prod()
        return ret - 1

    def get_attr_allocations(
        self, allocations: pd.DataFrame, attr: str, attr_weights: Dict[str, Dict[str, float]]
    ) -> pd.DataFrame:
        """
        Creates an allocations dataframe for a given meta attribute, where each
        column represents an attribute's value and each row contains the total amount
        allocated to that value.
        An extra column "unknown" is created to fit all holdings (or portions of
        holdings) that do not specify any allocation for this attribute.

        The arg attr_weights is used to specify holdings allocations with respect
        of this attribute's values, e.g.:
            {
                "VALUE_1": {"HOLDING_1": 0.5, "HOLDING_2": 0.9},
                "VALUE_2": {"HOLDING_1": 0.5}
            }

        For a single day X, assuming an equal price for HOLDING_1 and HOLDING_2 and
        an equal allocation of 100$ each, the resulting dataframe will look like
        this:

                  |VALUE_1 | VALUE_2 | unknown
                --+--------+---------+--------
                X |  140   |   50    |   10
        """

        columns = list(attr_weights) + ["unknown"]
        df = pd.DataFrame(0, columns=columns, index=allocations.index, dtype=np.float64)

        # Keep record of how much we allocate for each holding
        tot_holdings = defaultdict(lambda: 0)

        # Compute holdings allocations
        for entry, values in attr_weights.items():
            for holding_key, portion in values.items():
                df[entry] += portion * allocations[holding_key]
                tot_holdings[holding_key] += portion

        # Cumpute unknown allocations
        for holding_key in self.holdings_keys:
            allocation = tot_holdings[holding_key]

            # Floating point operations can lack of precision
            # so we should be slightly tollerant
            eps = 0.00001

            if allocation > 1 + eps:
                raise ValueError(
                    f"Attribute \"{attr}\" has more than 100% "
                    f"allocation for {holding_key}"
                )

            elif allocation > 1 - eps:
                continue

            df["unknown"] += allocations[holding_key] * (1 - allocation)
        return df

    def get_attr_earnings(
        self,
        attr_allocations: pd.DataFrame,
        transactions: List[Transaction],
        attr_weights: Dict[str, Dict[str, float]],
        ndays: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        Compute earnings by attribute allocations. Earnings are computed by
        calculating and ditributing the earnings for each holding to its
        attribute allocation. For example, if holding X has earned $100 and its sector
        allocation is 50% healthcare and 50% tech, then healthcare and tech sector
        have both produced $50 of earnings.
        Buy, sell and vest transaction are discounted from the earnings.
        """
        # Map holdings to their weights with respect of attributes
        holdings_alloc = defaultdict(lambda: {})

        for entry, values in attr_weights.items():
            for holding_key, portion in values.items():
                holdings_alloc[holding_key][entry] = portion

        # Earning for each attribute
        earnings = attr_allocations.copy()

        def _apply_deltas(date, delta, holding):
            weights_sum = 0
            for attr, weight in holdings_alloc[holding].items():
                earnings.loc[date:, attr] += delta * weight
                weights_sum += weight
            earnings.loc[date:, "unknown"] += delta * (1 - weights_sum)
            return earnings

        for trs in transactions:
            # Discount allocation increases after BUY
            if trs.action == TransactionAction.BUY:
                delta = -trs.amount.normalize_currency(self.conv_rates)
                earnings = _apply_deltas(trs.date, delta, trs.get_holding_key())

            # Discount allocation decreases after SELL
            elif trs.action == TransactionAction.SELL:
                delta = trs.amount.normalize_currency(self.conv_rates)
                earnings = _apply_deltas(trs.date, delta, trs.get_holding_key())

            # Discount vested stock (as it is not earning from investiment)
            elif trs.action == TransactionAction.VEST:
                holding = trs.get_holding_key()
                price = self.prices.loc[trs.date :][holding].iloc[0]
                holding_cur = self.holdings_currencies[holding]
                delta = -Price(
                    currency=holding_cur, amount=trs.quantity * price
                ).normalize_currency(self.conv_rates)
                earnings = _apply_deltas(trs.date, delta, holding)

        # Take only the last n days
        if ndays is not None:
            # pylint: disable=invalid-unary-operand-type
            earnings = earnings.iloc[-ndays:]

        # Subtract initial value so that we always start from zero
        earnings = earnings.add(-earnings.iloc[0])

        return earnings



"""
Config
"""

from typing import List, Dict, Optional, Generator, Any
from collections import defaultdict
from enum import Enum
import os
import csv
import datetime
import yaml
import pandas as pd
import numpy as np
import dateutil.parser
from .price import Price, Currency
from .transaction import Transaction, TransactionAction
from .holding import Holding

class ConfKeys(Enum):
    OPTIONS = "options"
    PRICES = "prices"
    META = "meta"
    TRANSACTIONS = "transactions"
    INCLUDE = "include"



class Config:
    """ Utility class representing a yaml project config """

    def __init__(self, cfg: str, path=None):
        self._cfg = yaml.load(cfg, Loader=yaml.SafeLoader)
        self._transactions = None
        self._prices = None
        self._holding_to_currency = {}
        self._files_provided = {}
        self._path = path

        self._load_includes()

    def _load_includes(self):
        for path in (self._cfg.get(ConfKeys.INCLUDE.value) or []):
            cfg_str = self._read_file(path=path)
            cfg = Config(cfg=cfg_str, path=path)
            self.merge(cfg=cfg)

    @staticmethod
    def from_file(path: str):
        """ Create Config object from file """
        with open(path) as fd:
            return Config(cfg=fd.read(), path=path)

    def merge(self, cfg: "Config"):
        self._cfg[ConfKeys.OPTIONS.value] = {
            **(cfg._cfg.get(ConfKeys.OPTIONS.value) or {}),
            **(self._cfg.get(ConfKeys.OPTIONS.value) or {}),
        }

        self._cfg[ConfKeys.META.value] = [
            *(cfg._cfg.get(ConfKeys.META.value) or []),
            *(self._cfg.get(ConfKeys.META.value) or []),
        ]

        self._cfg[ConfKeys.TRANSACTIONS.value] = [
            *(cfg._cfg.get(ConfKeys.TRANSACTIONS.value) or []),
            *(self._cfg.get(ConfKeys.TRANSACTIONS.value) or []),
        ]

        self._cfg[ConfKeys.PRICES.value] = [
            *(cfg._cfg.get(ConfKeys.PRICES.value) or []),
            *(self._cfg.get(ConfKeys.PRICES.value) or []),
        ]



    def provide_file(self, path: str, content: str):
        """
        Use this method if instead of reading from a path
        contained in the config you would like to provide
        directly the content of the file.
        This is similar to mocking.
        """
        self._files_provided[path] = content

    def _read_file(self, path: str) -> str:
        if not os.path.isabs(path) and self._path is not None:
            path = os.path.join(os.path.dirname(self._path), path)

        if path in self._files_provided:
            return self._files_provided[path]
        with open(path) as fd:
            return fd.read()

    def _get_opt(self, name: str) -> Any:
        opt = {}
        if ConfKeys.OPTIONS.value in self._cfg:
            opt = self._cfg[ConfKeys.OPTIONS.value]
        return opt.get(name)

    @property
    def title(self) -> int:
        """ Title of the project """
        return self._get_opt("title") or "Unnamed Project"

    @property
    def days(self) -> int:
        """ Number of days to analyse """
        return self._get_opt("days") or 90

    @property
    def end_date(self) -> datetime.datetime:
        """ When to stop the analysis """
        date_opt = self._get_opt("end_date")
        if date_opt is None:
            return datetime.datetime.now()

        return dateutil.parser.parse(date_opt, dayfirst=True)

    @property
    def start_date(self) -> datetime.datetime:
        """ When to start the analysis """
        return self.end_date - datetime.timedelta(days=self.days)

    @property
    def currency(self) -> Currency:
        """ Base currency to use """
        curr = self._get_opt("currency")
        if curr is None:
            return Currency.USD
        return Currency[curr]

    @property
    def transactions(self) -> List[Transaction]:
        """ All transactions ever done (random order) """
        if self._transactions is None:
            self._transactions = self._load_transactions()
        return self._transactions

    def transactions_by_holding(
        self, holding: Holding
    ) -> Generator[Transaction, None, None]:
        """ All transactions ever done for the given holding """
        for trs in self.transactions:
            if holding.match_transaction(transaction=trs):
                yield trs

    def _match_holding(self, match_section: Dict, holding: Holding) -> bool:
        if not isinstance(match_section, dict):
            raise ValueError("Expected a matching list")

        h = holding
        m = match_section
        t = "ticker"
        i = "isin"
        n = "name"

        return (
            (m.get(t) is None or m[t].strip() == getattr(h, t))
            and (m.get(i) is None or m[i].strip() == getattr(h, i))
            and (m.get(n) is None or m[n].strip() == getattr(h, n))
        )

    def _find_matching_entry(self, config_section: List[Dict], holding: Holding):
        for entry in config_section:
            if not isinstance(entry, dict) or "match" not in entry:
                raise ValueError("Expected a matching list")
            if not isinstance(entry["match"], dict):
                raise ValueError("A match section must contain key,value pairs")

            if self._match_holding(match_section=entry["match"], holding=holding):
                return entry

    def _get_prices_file(self, holding: Holding) -> Optional[str]:
        # If prices are provided use those (missing values are interpolated)
        if "prices" not in self._cfg:
            return

        if not isinstance(self._cfg["prices"], list):
            raise ValueError("Prices section must contain a matching list")

        match = self._find_matching_entry(
            config_section=self._cfg["prices"], holding=holding
        )

        if match is None:
            return

        if "file" not in match:
            raise ValueError("Prices entries must contain a file field")

        return match["file"]

    def _find_matching_holding(
        self, match_section: Dict, holdings: List[Holding]
    ) -> Holding:
        if not isinstance(match_section, dict):
            raise ValueError("Expected a matching list")

        for holding in holdings:
            if self._match_holding(match_section=match_section, holding=holding):
                return holding

    def _get_meta_attributes_apply(
        self, attrs: Dict, holding: Holding, apply: Dict
    ) -> defaultdict:
        holding_key = holding.get_key()

        for attr, val in apply.items():

            # Apply has the precendence over everything else.
            # If this attribute was already set, remove it.
            for holdings_alloc in attrs[attr].values():
                if holding_key in holdings_alloc:
                    del holdings_alloc[holding_key]

            if isinstance(val, str):
                attrs[attr][val][holding_key] = 1

            elif isinstance(val, dict):
                for k, v in val.items():
                    v = float(v.strip("%")) / 100.0
                    attrs[attr][k][holding_key] = v

    def _get_meta_attributes_composition(
        self, attrs: Dict, holding: Holding, holdings: List[Holding], composition: Dict
    ) -> defaultdict:
        # Collect all sub holdings
        sub_holdings = {}

        def _collect_by(field: str):
            if field not in composition:
                return

            if not isinstance(composition[field], dict):
                raise ValueError("Components must be key value pairs")

            for val, percentage in composition[field].items():

                # Try to find holding among known holdings
                sub_holding = self._find_matching_holding(
                    {field: val},
                    holdings,
                )

                if sub_holding is None:
                    sub_holding = self._find_matching_holding(
                        {field: val},
                        self._collect_holdings_from_meta(),
                    )

                if sub_holding is None:
                    continue

                sub_holdings[sub_holding.get_key()] = {
                    "holding": sub_holding,
                    "percentage": float(percentage.strip("%")) / 100.0,
                }

        _collect_by("name")
        _collect_by("ticker")
        _collect_by("isin")

        sub_holding_attrs = self.get_meta_attributes(
            [s["holding"] for s in sub_holdings.values()]
        )
        for attr, entries in sub_holding_attrs.items():
            for entry, holdings_alloc in entries.items():
                for holding_key, sub_alloc_percentage in holdings_alloc.items():
                    old_val = attrs[attr][entry].get(holding.get_key()) or 0.0
                    attrs[attr][entry][holding.get_key()] = old_val + (
                        sub_holdings[holding_key]["percentage"] * sub_alloc_percentage
                    )

    def _collect_holdings_from_meta(self) -> List[Holding]:
        holdings = []
        meta = self._cfg.get("meta") or []

        for entry in meta:
            m = entry["match"]
            name, ticker, isin = None, None, None

            if "name" in m:
                name = m["name"]

            if "ticker" in m:
                ticker = m["ticker"]

            if "isin" in m:
                isin = m["isin"]

            holdings.append(Holding(name=name, ticker=ticker, isin=isin))
        return holdings

    def get_meta_attributes(self, holdings: List[Holding]) -> defaultdict:
        """
        Gets all meta attributes for the given list of holdings
        The returned attrs dict is structured as follows:
          attrs[<attribute name>][<entry>][<holding key>] = x
        where 0 < x <= 1.

        For instance, a meta attr "type: equity" will result in:
          attrs["type"]["equity"][<holding key>] = 1
        """
        attrs = defaultdict(lambda: defaultdict(dict))

        meta = self._cfg.get("meta")
        if meta is None:
            return attrs

        for entry in meta:
            if not isinstance(entry, dict) or "match" not in entry:
                raise ValueError("Expected a matching list")

            holding = self._find_matching_holding(entry["match"], holdings=holdings)
            if holding is None:
                continue

            if "composition" in entry:
                self._get_meta_attributes_composition(
                    attrs=attrs,
                    holding=holding,
                    holdings=holdings,
                    composition=entry["composition"],
                )

            if "apply" in entry:
                self._get_meta_attributes_apply(
                    attrs=attrs, holding=holding, apply=entry["apply"]
                )

        return attrs

    def get_currency(self, holding: Holding) -> Optional[Currency]:
        """ Currency for the given holding (if provided) """
        path = self._get_prices_file(holding=holding)
        if path is None:
            return

        for row in csv.DictReader(self._read_file(path)):
            price = Price.from_str(price=row["price"])
            return price.currency

    def get_prices(
        self,
        holding: Holding,
        start: datetime.datetime,
        end: Optional[datetime.datetime] = None,
    ) -> Optional[pd.DataFrame]:
        """ Prices history for the given holding (if provided) """

        path = self._get_prices_file(holding=holding)
        if path is None:
            return

        if end is None:
            end = self.end_date

        index = pd.date_range(start=start, end=end, freq="D")
        prices = pd.Series(index=index, dtype=np.float64)
        prices.name = holding.get_key()

        for row in csv.DictReader(self._read_file(path).split("\n")):
            date = dateutil.parser.parse(row["date"], dayfirst=True)
            price = Price.from_str(price=row["price"])

            if date >= start and date <= end:
                prices[date] = price.amount

        return prices

    def _load_transactions_schwab(self, filename: str) -> List[Transaction]:
        trs = []

        csvfile = self._read_file(filename).split("\n")

        # skip first line, which contains the document title
        csvfile = csvfile[1:]

        # skip last line, which contains the total
        for row in list(csv.DictReader(csvfile))[:-1]:
            # dates are sometime expressed as "mm/dd/yyyy as of mm/dd/yyy"
            # in which case we pick the first date
            date_str = row["Date"].split(" ")[0]
            date = datetime.datetime.strptime(date_str, "%m/%d/%Y")

            action = TransactionAction.from_schwab_action(row["Action"])

            price_str = row["Price"]
            price = Price.from_str(price_str) if price_str else None

            fees_str = row["Fees & Comm"]
            fees = Price.from_str(fees_str) if fees_str else None

            quantity_str = row["Quantity"]
            quantity = float(quantity_str) if quantity_str else None

            amount_str = row["Amount"]
            if amount_str:
                if action == TransactionAction.TAX or action == TransactionAction.BUY:
                    amount = Price.from_str(amount_str, expect_negative=True)
                else:
                    amount = Price.from_str(amount_str)
            else:
                amount = None

            trs.append(
                Transaction(
                    action=action,
                    date=date,
                    ticker=row["Symbol"] or None,
                    name=row["Description"] or None,
                    price=price,
                    quantity=quantity,
                    fees=fees,
                    amount=amount,
                )
            )
        return trs

    def _load_transactions_standard(self, filename: str) -> List[Transaction]:
        trs = []
        for row in list(csv.DictReader(self._read_file(filename).split("\n"))):
            date = dateutil.parser.parse(row["date"], dayfirst=True)
            action = TransactionAction(row["action"])
            price_str = row["price"]
            price = Price.from_str(price_str) if price_str else None

            fees_str = row["fees"]
            fees = Price.from_str(fees_str) if fees_str else None

            quantity_str = row["quantity"]
            quantity = float(quantity_str) if quantity_str else None

            amount_str = row["amount"]
            amount = Price.from_str(amount_str) if amount_str else None

            trs.append(
                Transaction(
                    action=action,
                    date=date,
                    ticker=row["ticker"] or None,
                    name=row["name"] or None,
                    price=price,
                    quantity=quantity,
                    fees=fees,
                    amount=amount,
                )
            )
        return trs

    def _load_transactions(self) -> List[Transaction]:
        transactions = []

        if "transactions" not in self._cfg:
            return transactions

        for entry in self._cfg["transactions"]:
            loader = entry["format"]
            if loader == "standard":
                transactions.extend(
                    self._load_transactions_standard(filename=entry["file"])
                )
            elif loader == "schwab":
                transactions.extend(
                    self._load_transactions_schwab(filename=entry["file"])
                )
            else:
                raise ValueError(f"Unsupported transactions' format {loader}")

        transactions = filter(lambda trs: trs.date <= self.end_date, transactions)
        return list(sorted(transactions, key=lambda trs: trs.date))



import os
from shutil import copyfile
import click
from .project import Project

@click.group()
def main():
    pass

@main.command("gen-report")
@click.argument("config")
@click.argument("dest")
def make_report(config: str, dest: str):
    """
    Generate an html report from a config
    """
    proj = Project(config=config)
    proj.gen_report(dest)

@main.command("new-project")
@click.argument("dest")
def new_project(dest: str):
    """
    Generate a template for a new project
    """
    src = os.path.join(os.path.dirname(__file__), "project_template")
    files = ["project.yml", "transactions.csv"]
    for f in files:
        copyfile(os.path.join(src,f), os.path.join(dest,f))
    click.echo("New project created !")
    click.echo("To generate a report run:")
    click.echo("  inverno gen-report project.yml DEST")

    



from typing import Dict
from datetime import datetime
import shutil
import tempfile
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as ani
import yfinance as yf
from currency_converter import CurrencyConverter
from jinja2 import Environment, PackageLoader
from .balance import Balance
from .price import Currency, Price
from .common import log_info, log_warning
from .holding import Holding
from .analysis import Analysis
from .config import Config


class Project:
    """
    Root class for handling a project and the creation of a report
    """

    def __init__(self, config: str):

        self.cfg = Config.from_file(path=config)

        # Conversion rates to the dest currency
        self._dst_currency_rates = self._get_currency_rates(self.cfg.currency.name)

        # Balances after each transaction (max one balance per day)
        self.balances = Balance.get_balances(transactions=self.cfg.transactions)

        # For each holding identity (key) this list contains the
        # first one ever hold
        self._first_holdings = self._get_first_holdings()

        # Maps holdings to their currency
        self._holding_to_currency = self._get_currencies()

        # Daily prices for every holding
        self._prices = self._get_prices()

        # All meta attributes
        self._meta = self.cfg.get_meta_attributes(
            [h["holding"] for h in self._first_holdings.values()]
        )

        # Adding dummy attribute holdings
        self._meta["holdings"] = {
            h["holding"].get_key(): {h["holding"].get_key(): 1.0}
            for h in self._first_holdings.values()
        }

    def _get_currency_rates(self, currency: str) -> Dict[str, float]:
        cc = CurrencyConverter()
        rates = {currency: 1.0}
        for cu in cc.currencies:
            try:
                rates[cu] = cc.convert(1, currency, cu)
            except Exception:
                pass
        return rates

    def _get_attrs_report_data(self, analysis: Analysis, allocations: pd.DataFrame):
        reports = []
        for attr in self._meta:
            log_info(f"Generating report for attribute {attr}")
            reports.append({"name": attr.capitalize(), "reports": []})

            attr_alloc = analysis.get_attr_allocations(
                allocations=allocations, attr=attr, attr_weights=self._meta[attr],
            )

            # Get current allocation from last (more recent) row
            last_alloc = attr_alloc.tail(1).values.tolist()[0]
            reports[-1]["reports"].append(
                {
                    "type": "piechart",
                    "name": "Allocation",
                    "data": last_alloc,
                    "labels": list(attr_alloc.columns),
                    "help": "Shows how much is allocated for each type in the portfolio.",
                }
            )

            # Allocation history
            reports[-1]["reports"].append(
                {
                    "type": "areachart_stacked",
                    "name": "Allocation history",
                    "datasets": [
                        {"label": c, "data": attr_alloc[c].tolist()}
                        for c in attr_alloc.columns
                    ],
                    "labels": [d.strftime("%d %b %Y") for d in attr_alloc.index],
                    "show_legend": True,
                    "help": "Full history of allocations.",
                }
            )

            # Earnings
            earnings = analysis.get_attr_earnings(
                attr_allocations=attr_alloc,
                transactions=self.cfg.transactions,
                attr_weights=self._meta[attr],
            )
            earnings_perc = (earnings / attr_alloc) * 100
            earnings_perc.replace([np.inf, -np.inf], np.nan, inplace=True)
            earnings_perc = earnings_perc.fillna(0.0)

            reports[-1]["reports"].append(
                {
                    "type": "multi",
                    "name": "Earnings",
                    "help": "Earnings growth for each type adjusted for cash flow.",
                    "reports": [
                        {
                            "type": "areachart",
                            "name": "Value",
                            "datasets": [
                                {"label": c, "data": earnings[c].tolist()}
                                for c in earnings.columns
                            ],
                            "labels": [d.strftime("%d %b %Y") for d in earnings.index],
                            "show_legend": True,
                        },
                        {
                            "type": "areachart",
                            "name": "Percentage (%)",
                            "datasets": [
                                {"label": c, "data": earnings_perc[c].tolist()}
                                for c in earnings_perc.columns
                            ],
                            "labels": [
                                d.strftime("%d %b %Y") for d in earnings_perc.index
                            ],
                            "show_legend": True,
                            "format": "percent",
                        },
                    ],
                }
            )

        return sorted(reports, key=lambda r: r["name"].lower() != "holdings")

    def _get_report_data(self):
        analysis = Analysis(
            prices=self._prices,
            conv_rates=self._dst_currency_rates,
            holdings_currencies=self._holding_to_currency,
        )

        # Balances graph
        allocations = analysis.get_allocations(
            balances=self.balances.values(), ndays=self.cfg.days
        )
        balances = {
            "datasets": [
                {
                    "label": "Balance",
                    "data": allocations.sum(axis=1).values.tolist(),
                },
            ],
            "labels": [d.strftime("%d %b %Y") for d in allocations.index],
        }

        # Earning graph
        earnings_s = analysis.get_earnings(
            allocations=allocations,
            transactions=self.cfg.transactions,
            ndays=self.cfg.days,
        )

        benchmarks = self._get_benchmarks(
            start=self.cfg.start_date,
            end=self.cfg.end_date,
        )
        earnings = {
            "datasets": [
                {
                    "label": "Earnings",
                    "data": earnings_s.values.tolist(),

                },
                *[ { "label": k, "data": v.values.tolist()} for k,v in benchmarks.items() ]
            ],
            "labels": [d.strftime("%d %b %Y") for d in earnings_s.index],
        }

        # Generate report data for all known attributes
        attrs_report = self._get_attrs_report_data(
            analysis=analysis, allocations=allocations
        )

        # Rate of return
        ror = analysis.get_ror(allocations, earnings_s)

        # Current number of holdings
        nb_holdings = (allocations.iloc[-1] > 0).sum()

        return {
            "balances": balances,
            "earnings": earnings,
            "attrs": attrs_report,
            "ror": ror,
            "nb_holdings": nb_holdings,
        }

    def gen_report(self, dst: str):
        """ Create an html report at the given destination """

        # First we copy the html dir into a tmp folder
        src = os.path.join(os.path.dirname(__file__), "html")
        tmp_dst = tempfile.TemporaryDirectory(suffix="." + __package__)
        shutil.copytree(src=src, dst=tmp_dst.name, dirs_exist_ok=True)

        # This is the data that we will feed to the report
        report_data = self._get_report_data()

        # Generate report
        jinja_env = Environment(
            loader=PackageLoader("inverno", "html"),
            # autoescape=select_autoescape(["html", "xml"]),
            autoescape=True,
        )

        index_path = os.path.join(tmp_dst.name, "index.html")
        index_template = jinja_env.get_template(name="index.html")
        index_template.stream(
            cfg=self.cfg,
            attrs=report_data["attrs"],
            balance=Price(
                currency=self.cfg.currency,
                amount=report_data["balances"]["datasets"][0]["data"][-1],
            ).to_string(),
            earnings_amount=Price(
                currency=self.cfg.currency,
                amount=report_data["earnings"]["datasets"][0]["data"][-1],
            ).to_string(),
            ror=f"{report_data['ror']*100: .2f}",
            nb_holdings=report_data["nb_holdings"],
            balances=report_data["balances"],
            earnings=report_data["earnings"],
            transactions=sorted(
                self.cfg.transactions, key=lambda t: t.date, reverse=True
            ),
        ).dump(index_path)

        # Move report to dst
        if os.path.isdir(dst):
            dst = os.path.join(dst, os.path.basename(dst))
        shutil.move(src=tmp_dst.name, dst=dst)

    def _gen_animated_plot(self, df: pd.DataFrame, dst: str):
        fig = plt.figure()
        plt.xticks(rotation=45, ha="right", rotation_mode="anchor")
        plt.subplots_adjust(bottom=0.2, top=0.9)
        colors = [
            "red",
            "green",
            "blue",
            "orange",
            "black",
            "violet",
            "brown",
            "yellow",
            "purple",
        ]

        frames = min(self.cfg.days, df.index.size)
        start = df.index.size - frames

        def update_frame(i):
            plt.legend(df.columns)
            p = plt.plot(df[start : start + i].index, df[start : start + i].values)
            for i in range(len(df.columns)):
                p[i].set_color(colors[i % len(colors)])

        animator = ani.FuncAnimation(fig, update_frame, frames=frames, interval=300)
        animator.save(dst, fps=10)

    def _get_first_holdings(self):
        # Collect holdngs and earliest date
        holdings = {}
        for balance in self.balances.values():
            for holding in balance.holdings.values():
                if holding.get_key() not in holdings:
                    holdings[holding.get_key()] = {
                        "date": balance.date,
                        "holding": holding,
                    }
        return holdings

    def _get_prices(self):
        prices = []
        end_date = self.cfg.end_date
        for _, entry in self._first_holdings.items():
            price_history = self._get_holding_prices(
                start=entry["date"], end=end_date, holding=entry["holding"]
            )
            if price_history is not None:
                if all([np.isnan(p) for p in price_history.tail(7)]):
                    log_warning(
                        "Most recent price is older than one "
                        f"week for {entry['holding'].get_key()}"
                    )
                prices.append(price_history)

        # Put all together in a single dataframe
        prices = pd.concat(prices, axis=1, join="outer")

        # Use linear interpolation to cover NaNs
        prices = prices.interpolate(method="time", axis=0, limit_direction="both")

        return prices

    def _get_currencies(self):
        currencies = {}
        for entry in self._first_holdings.values():
            holding = entry["holding"]
            log_info(f"Getting currency for {holding.get_key()}")

            # Try from transactions
            for trs in self.cfg.transactions_by_holding(holding=holding):
                if trs.price is not None:
                    currencies[holding.get_key()] = trs.price.currency
                    break

            if holding.get_key() in currencies:
                continue

            # Try from prices
            c = self.cfg.get_currency(holding=holding)
            if c is not None:
                currencies[holding.get_key()] = c
                continue

            # Try from Yahoo Finance
            if holding.ticker is not None:
                ticker = yf.Ticker(holding.ticker)
                try:
                    c = Currency[ticker.info["currency"]]
                    currencies[holding.get_key()] = c
                    continue
                except KeyError:
                    pass

            raise ValueError(f"Couldn't determine currency for {holding.get_key()}")

        return currencies

    def _get_benchmarks(self, start: datetime, end: datetime):
        def _reindex(s: pd.Series):
            s = s.add(-s.iloc[0])
            return s.reindex(
                index=pd.date_range(s.index.min(), s.index.max()),
                method="pad",
            )

        benchmarks = {}

        # S&P 500
        ticker = yf.Ticker("^GSPC")
        prices = ticker.history(start=start, end=end, interval="1d")["Close"]
        if prices.size > 0:
            benchmarks["S&P 500"] = _reindex(prices)

        # DJIA
        ticker = yf.Ticker("^DJI")
        prices = ticker.history(start=start, end=end, interval="1d")["Close"]
        if prices.size > 0:
            benchmarks["DJIA"] = _reindex(prices)
            

        # NASDAQ
        ticker = yf.Ticker("^IXIC")
        prices = ticker.history(start=start, end=end, interval="1d")["Close"]
        if prices.size > 0:
            benchmarks["NASDAQ"] = _reindex(prices)

        # Russel 2000 
        ticker = yf.Ticker("^RUT")
        prices = ticker.history(start=start, end=end, interval="1d")["Close"]
        if prices.size > 0:
            benchmarks["Russell 2000"] = _reindex(prices)

        return benchmarks

    def _get_holding_prices(
        self, start: datetime, end: datetime, holding: Holding
    ) -> pd.Series:
        def _reindex(s: pd.Series):
            return s.reindex(
                index=pd.date_range(s.index.min(), s.index.max()),
                method="pad",
            )

        prices = self.cfg.get_prices(holding=holding, start=start, end=end)
        if prices is not None:
            log_info(f"Using user-provided prices for {holding.get_key()}")
            return _reindex(prices)

        # Try to fetch prices from Yahoo Finance
        if holding.ticker is not None:
            ticker = yf.Ticker(holding.ticker)
            prices = ticker.history(start=start, end=end, interval="1d")["Close"]
            prices.name = holding.get_key()
            if prices.size > 0:
                log_info(f"Using Yahoo Finance prices for {holding.get_key()}")
                return _reindex(prices)

        # Try to infer from transaction data
        log_warning(
            f"Inferring prices from transactions for {holding.get_key()}"
            ", prices could be inaccurate"
        )
        index = pd.date_range(start=start, end=end, freq="D")
        prices = pd.Series(index=index, dtype=np.float64)
        prices.name = holding.get_key()

        for trs in self.cfg.transactions_by_holding(holding=holding):
            if trs.price is None:
                continue
            prices[trs.date] = trs.price.amount

        return _reindex(prices)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inverno']

package_data = \
{'': ['*'],
 'inverno': ['html/*',
             'html/css/*',
             'html/js/*',
             'html/js/demo/*',
             'html/js/src/*',
             'html/vendor/bootstrap/js/*',
             'html/vendor/bootstrap/scss/*',
             'html/vendor/bootstrap/scss/mixins/*',
             'html/vendor/bootstrap/scss/utilities/*',
             'html/vendor/bootstrap/scss/vendor/*',
             'html/vendor/chart.js/*',
             'html/vendor/datatables/*',
             'html/vendor/fontawesome-free/*',
             'html/vendor/fontawesome-free/css/*',
             'html/vendor/fontawesome-free/js/*',
             'html/vendor/fontawesome-free/less/*',
             'html/vendor/fontawesome-free/metadata/*',
             'html/vendor/fontawesome-free/scss/*',
             'html/vendor/fontawesome-free/sprites/*',
             'html/vendor/fontawesome-free/svgs/brands/*',
             'html/vendor/fontawesome-free/svgs/regular/*',
             'html/vendor/fontawesome-free/svgs/solid/*',
             'html/vendor/fontawesome-free/webfonts/*',
             'html/vendor/jquery-easing/*',
             'html/vendor/jquery/*',
             'project_template/*']}

install_requires = \
['CurrencyConverter>=0.16.1,<0.17.0',
 'Jinja2>=2.11.3,<3.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'click>=7.1.2,<8.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'pandas>=1.2.3,<2.0.0',
 'yfinance>=0.1.59,<0.2.0']

entry_points = \
{'console_scripts': ['inverno = inverno.cli:main']}

setup_kwargs = {
    'name': 'inverno',
    'version': '0.1.4',
    'description': 'Investments portfolio tracking tool',
    'long_description': None,
    'author': 'werew',
    'author_email': 'werew@ret2libc.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Union
from .price import Price
from .transaction import Transaction

class Holding:
    def __init__(
        self,
        quantity: Optional[float] = None,
        name: Optional[str] = None,
        ticker: Optional[str] = None,
        isin: Optional[str] = None,
    ):
        self.quantity = quantity
        self.name = None if name is None else name.strip()
        self.ticker = None if ticker is None else ticker.strip()
        self.isin = None if isin is None else isin.strip()

    def get_key(self):
        # prioritize stronger identifiers
        if self.isin is not None:
            return self.isin
        elif self.ticker is not None:
            return self.ticker
        elif self.name is not None:
            return self.name
        raise ValueError("Cannot construct holding name")

    def __hash__(self):
        return hash(self.get_key())

    def match_transaction(self, transaction: Transaction) -> bool:
        if self.ticker is not None and self.ticker != transaction.ticker:
            return False

        if self.isin is not None and self.isin != transaction.isin:
            return False

        if self.name is not None and self.name != transaction.name:
            return False

        return True

    def __repr__(self):
        return f"{self.quantity} {self.get_key()}"

    def __add__(self, other: Union["Holding", Price, float]):
        new_quantity = self.quantity

        if isinstance(other, float):
            new_quantity += other

        elif isinstance(other, Holding):
            if self.get_key() != other.get_key():
                raise ValueError(
                    "Trying to add up holding with different keys:"
                    f" {self.get_key()} {other.get_key()}"
                )
            new_quantity += other.quantity
        else:
            raise ValueError(f"Incompatible type {type(other)}")

        return Holding(
            quantity=new_quantity,
            name=self.name,
            ticker=self.ticker,
            isin=self.isin,
        )

    def __sub__(self, other: Union["Holding", Price, float]):
        new_quantity = self.quantity

        if isinstance(other, float):
            new_quantity -= other

        elif isinstance(other, Holding):
            if self.get_key() != other.get_key():
                raise ValueError(
                    "Trying to add up holding with different keys:"
                    f" {self.get_key()} {other.get_key()}"
                )
            new_quantity -= other.quantity
        else:
            raise ValueError(f"Incompatible type {type(other)}")

        return Holding(
            quantity=new_quantity,
            name=self.name,
            ticker=self.ticker,
            isin=self.isin,
        )

    def __neg__(self):
        return Holding(
            quantity=-self.quantity,
            name=self.name,
            ticker=self.ticker,
            isin=self.isin,
        )


from typing import Optional, Dict, List
import copy
from datetime import datetime
from .transaction import Transaction, TransactionAction
from .price import Currency
from .holding import Holding


class Balance:
    def __init__(
        self,
        date: datetime,
        holdings: Optional[Dict[str, Holding]] = None,
        cash: Optional[Dict[Currency, float]] = None,
    ):
        self.date = date
        self.holdings = holdings or {}
        self.cash = cash or {}

    def process_transaction(self, transaction: Transaction) -> "Balance":
        new_balance = copy.deepcopy(self)
        new_balance.date = max(transaction.date, new_balance.date)

        if transaction.action == TransactionAction.BUY:
            self._process_buy_transaction(
                new_balance=new_balance, transaction=transaction
            )
        elif transaction.action == TransactionAction.SELL:
            self._process_sell_transaction(
                new_balance=new_balance, transaction=transaction
            )
        elif transaction.action == TransactionAction.CASH_IN:
            self._process_cash_transaction(
                new_balance=new_balance, transaction=transaction, out=False,
            )
        elif transaction.action == TransactionAction.CASH_OUT:
            self._process_cash_transaction(
                new_balance=new_balance, transaction=transaction, out=True,
            )
        elif transaction.action == TransactionAction.DIV:
            self._process_div_transaction(
                new_balance=new_balance, transaction=transaction
            )
        elif transaction.action == TransactionAction.TAX:
            self._process_tax_transaction(
                new_balance=new_balance, transaction=transaction
            )
        elif transaction.action == TransactionAction.VEST:
            self._process_vest_transaction(
                new_balance=new_balance, transaction=transaction
            )
        else:
            raise ValueError(
                f"Couldn't process transaction action {transaction.action.name}:"
                " this action is currently not fully supported"
            )

        return new_balance

    def _make_holding(self, transaction: Transaction) -> Holding:
        return Holding(
            quantity=transaction.quantity,
            name=transaction.name,
            ticker=transaction.ticker,
            isin=transaction.isin,
        )

    def get_cash_balance(self, currency: Currency) -> float:
        cash = self.cash.get(currency)
        if cash is None:
            cash = 0.0
        return cash

    def _process_sell_transaction(
        self, new_balance: "Balance", transaction: Transaction
    ):
        # Reduce available holdings
        new_holding = self._make_holding(transaction=transaction)
        if new_holding.get_key() in new_balance.holdings:
            new_balance.holdings[new_holding.get_key()] -= new_holding
        else:
            new_balance.holdings[new_holding.get_key()] = -new_holding

        # Update cash after transaction
        cash = new_balance.get_cash_balance(transaction.price.currency)
        new_cash = cash + (transaction.price * transaction.quantity).amount
        if transaction.fees:
            new_cash -= transaction.fees.amount
        new_balance.cash[transaction.price.currency] = new_cash

    def _process_buy_transaction(
        self, new_balance: "Balance", transaction: Transaction
    ):
        # Reduce cash for the given currency
        cash = new_balance.get_cash_balance(transaction.price.currency)
        new_cash = cash - (transaction.price * transaction.quantity).amount
        if transaction.fees:
            new_cash -= transaction.fees.amount
        new_balance.cash[transaction.price.currency] = new_cash

        # Add holdings
        new_holding = self._make_holding(transaction=transaction)
        if new_holding.get_key() in new_balance.holdings:
            new_balance.holdings[new_holding.get_key()] += new_holding
        else:
            new_balance.holdings[new_holding.get_key()] = new_holding

    def _process_cash_transaction(
        self, new_balance: "Balance", transaction: Transaction, out: bool,
    ):
        # Reduce cash for the given currency
        cash = new_balance.get_cash_balance(transaction.amount.currency)

        if out:
            new_cash = cash - transaction.amount.amount
        else:
            new_cash = cash + transaction.amount.amount

        if transaction.fees:
            new_cash -= transaction.fees.amount
        new_balance.cash[transaction.amount.currency] = new_cash

    def _process_div_transaction(
        self, new_balance: "Balance", transaction: Transaction
    ):
        self._process_cash_transaction(new_balance=new_balance, transaction=transaction, out=False)

    def _process_tax_transaction(
        self, new_balance: "Balance", transaction: Transaction
    ):
        self._process_cash_transaction(new_balance=new_balance, transaction=transaction, out=True)

    def _process_vest_transaction(
        self, new_balance: "Balance", transaction: Transaction
    ):
        # Apply fees (if any) (??)
        if transaction.fees:
            cash = new_balance.get_cash_balance(transaction.fees.currency)
            new_cash = cash - transaction.fees.amount
            new_balance.cash[transaction.fees.currency] = new_cash

        # Add holdings
        new_holding = self._make_holding(transaction=transaction)
        if new_holding.get_key() in new_balance.holdings:
            new_balance.holdings[new_holding.get_key()] += new_holding
        else:
            new_balance.holdings[new_holding.get_key()] = new_holding

    @staticmethod
    def get_balances(transactions: List[Transaction]) -> Dict[datetime, "Balance"]:
        """
        Process a sorted list of transactions and return all intermediary balances
        """
        if not transactions:
            return {}

        balances = {}

        start_date = transactions[0].date
        last_balance = Balance(date=start_date)

        for trs in transactions:
            last_balance = last_balance.process_transaction(trs)
            balances[last_balance.date] = last_balance

        return balances

    def __str__(self):
        return (
            f"CASH: {list(self.cash.values())}, "
            f"HOLDINGS: {list(self.holdings.values())}"
        )

    def __repr__(self):
        return self.__str__()
